/*
 ******************************************************************
 * History
 * 15-Oct-94  Jeff Shufelt (js), Carnegie Mellon University
 *      Prepared for 15-681, Fall 1994.
 *
 ******************************************************************
 */

#include <stdlib.h>
#include <stdio.h>
#include <malloc.h>
#include <string.h>
#include "image.h"
#include "pnm.h"

#define STR_SIZE 1024



/* 
 *
 * function definitions 
 *
 *
 */


/* return values: image = OK, NULL = problem */

img_t *img_read_ppm (char *filename)
{
  img_t     *new;
  FILE      *fptr;
  char       line[512];
  int        type, maxval, pix_val, in_val, parse_state;
  img_dim_t  wd, ht, pix_cnt, i;
  int        chan_cnt;
  pix_val_t *data;

  new = img_new();
  if (strcmp (filename, "stdin") == 0
      || strcmp (filename, "-") == 0)
    fptr = stdin;
  else if ((fptr = fopen(filename, "r")) == NULL) {
    fprintf(stderr,"IMGOPEN: Couldn't open '%s'\n", filename);
    return(NULL);
  }

  new->name = strdup (img_basename(filename));

  /*** Scan pnm type information, expecting P5 ***/
  fgets(line, 512, fptr);
  sscanf(line, "P%d", &type);
  switch (type) {
  case 2:
  case 5:
    chan_cnt = 1;
    new->ftype = IMG_FTYPE_PGM;
    break;
  case 3:
  case 6:
    chan_cnt = 3;
    new->ftype = IMG_FTYPE_PPM;
    break;
  default:
    fprintf(stderr,"IMGOPEN: cannot handle files of type P%d\n", type);
    fclose(fptr);
    return(NULL);
  }

  /*** Get dimensions of pgm ***/
  while (1) {
    fgets(line, 511, fptr);
    if (line[0] != '#') 
      break;
  }
  sscanf(line, "%ld %ld", &wd, &ht);
  new->height = ht;
  new->width = wd;
  new->chan_cnt = chan_cnt;
  new->stride = wd * chan_cnt;

  /*** Get maxval ***/
  while (1) {
    fgets(line, 511, fptr);
    if (line[0] != '#') 
      break;
  }
  sscanf(line, "%d", &maxval);
  if (maxval > 255) {
    fprintf(stderr,"IMGOPEN: Only handles pgm files of 8 bits or less\n");
    fclose(fptr);
    return(NULL);
  }

  pix_val = 0;
  pix_cnt = ht * wd * chan_cnt;
  data = (pix_val_t *) malloc (pix_cnt * sizeof(pix_val_t));
  new->data = data;
  if (data == NULL) {
    fprintf(stderr,"IMGOPEN: Couldn't allocate space for image data\n");
    fclose(fptr);
    return(NULL);
  }

  if (type == 5 || type == 6) {

    for (i = 0; i < pix_cnt; i++) {
      *data++ = fgetc(fptr);
    }

  }

  else if (type == 2 || type == 3) {

    for (i = 0; i < pix_cnt; i++) {

      parse_state = 0;

      while (parse_state < 2) {
	in_val = (int)fgetc(fptr) - (int)'0';
	switch (parse_state) {
	case 0:
	  if (in_val >= 0 && in_val <= 9) {
	    pix_val = in_val;
	    parse_state = 1;
	  }
	  break;
	default:
	  if (in_val < 0 || in_val > 9) 
	    parse_state = 2;
	  else
	    pix_val = pix_val * 10 + in_val;
	  break;
	}
      }
      *data++ = pix_val;
    }
  }
  
  else {
    fprintf(stderr,"IMGOPEN: Fatal impossible error\n");
    fclose(fptr);
    return (NULL);
  }

  fclose(fptr);
  return (new);
}


/* return values: image = OK, NULL = problem */

img_t *img_read_pgm (char *filename)
{
  return (img_read_ppm (filename));
}


/* output in binary ppm format */

/* return values: 0 = OK, 1 = problem */

int img_write_ppm (img_t *img, char *filename)
{
  img_dim_t ht, wd;
  FILE      *fptr;
  long       val_cnt;

  if (strcmp (filename, "stdout") == 0
      || strcmp (filename, "-") == 0)
    fptr = stdout;
  else
    fptr = fopen(filename, "w");
  if (fptr == NULL)
    return (1);
  if (img->chan_cnt == 1)
    fprintf(fptr, "P5\n");
  else
    fprintf(fptr, "P6\n");

  ht = img->height; 
  wd = img->width;
  fprintf(fptr, "%ld %ld\n", wd, ht);
  fprintf(fptr, "255\n");

  val_cnt = ht * wd * img->chan_cnt;
  if (fwrite (img->data, val_cnt, 1, fptr) != 1) {
    fprintf(fptr, "\n");
    fclose(fptr);
    return (1);		/* an error ocurred */
  }
  fprintf(fptr, "\n");
  fclose(fptr);
  return (0);
}


/* output in ascii ppm format */

/* return values: 0 = OK, 1 = problem */

int img_write_ppm_ascii (img_t *img, char *filename)
{
  img_dim_t ht, wd;
  FILE      *fptr;
  pix_val_t *data;
  long       i, val_cnt;

  if (strcmp (filename, "stdout") == 0
      || strcmp (filename, "-") == 0)
    fptr = stdout;
  else
    fptr = fopen(filename, "w");
  if (fptr == NULL)
    return (1);
  if (img->chan_cnt == 1)
    fprintf(fptr, "P2\n");
  else
    fprintf(fptr, "P3\n");

  ht = img->height; 
  wd = img->width;
  fprintf(fptr, "%ld %ld\n", wd, ht);
  fprintf(fptr, "255\n");

  data = img->data;
  val_cnt = ht * wd * img->chan_cnt;
  for (i = 0; i < val_cnt; i++) {
    fprintf(fptr, "%d ", (int)*data++);
      if (i % 15 == 14) 
        fprintf(fptr, "\n");
  }
  fprintf(fptr, "\n");
  fclose(fptr);
  return (0);
}


/* return values: 0 = OK, 1 = problem */

int img_write_pgm (img_t *img, char *filename)
{
  return (img_write_ppm (img, filename));
}


/* return values: 0 = OK, 1 = problem */

int imgl_read_pgm_images (img_list_t *image_list)
{
  img_t *image;
  int i;
  char *fname;
  
  for (i = 0; i < image_list->cnt; i++) {

    fname = image_list->fname[i];

    fprintf(stderr,"Loading '%s'...", fname);
    if ((image = img_read_pgm (fname)) == NULL) {
      fprintf(stderr,"Couldn't open '%s'\n", fname);
      return (1);
    }
   image_list->image[i] = image;

  }

  return (0);

}


