/*%********************************************************************
 *
 * File: util.h
 *
 * Author:  Charles Rosenberg 
 * Created: July 7, 1998
 * 
 * Header file for fnetutil.c
 *
 * General purpose utilities
 *
 * Modification History
 *
 * 07/07/98, cjr: created from fnetutil.h
 *
 **********************************************************************
 */

/* redundant include protection */

#ifndef UTIL_UTIL_H
#define UTIL_UTIL_H

/* constants */

#define FGETS_TEMP_STR_SIZE 16384
#define WSPACE_CHARS " \t\n\015"

/* data types */


/* function prototypes */

void *emalloc (size_t size, char *name);
void *emalloc2 (size_t size, char *src_name, int line, char *fun_name);
void efree (void *ptr);
void *erealloc(void *ptr, size_t size, char *name);
void *erealloc2 (void *ptr, size_t size, 
		 char *src_name, int line, char *fun_name);
char *estrdup (const char *s, char *name);
char *estrdup2 (const char *s, char *src_name, int line, char *fun_name);
FILE *efopen (char *path, char *mode, char *name);
FILE *efopen2 (char *path, char *mode, 
	       char *src_name, int line, char *fun_name);
char *fix_str (char *str);
int estrncasecmp (const char *s1, const char *s2);
char *snf (char *str);
int simple_regexp_cmp (char *srcstr, char *regexp);
char *find_substr (char *src_str, char *find_str);
long get_flen (char *fname);
int exist_dir (char *dname);
double drandom();
long lrandom (long n);
double norm_random ();
char *fgets_clean_eof(char *s, FILE *stream);
char *fgets_clean(char *s, FILE *stream);
char *estrsep(char **str);
void str_tolower (char *str);

/* macros */

#define EFOPEN(path,mode)  efopen2(path,mode,__FILE__,__LINE__,__FUNCTION__)
#define EMALLOC(size)      emalloc2(size,__FILE__,__LINE__,__FUNCTION__)
#define EREALLOC(ptr,size) erealloc2(ptr,size,__FILE__,__LINE__,__FUNCTION__)
#define ESTRDUP(s)	   estrdup2(s,__FILE__,__LINE__,__FUNCTION__) 
#define EFREE(ptr)         efree(ptr)

/* display a message to stderr about the error location */

#define ERRHERE()  fprintf(stderr, "\nError occured at: %s:%d %s()\n",__FILE__,__LINE__,__FUNCTION__)


/* a slightly cleaned up version of strncasecmp
 * automatically calculates the length of the second argument
 */

#define estrncasecmp(s1,s2) (strncasecmp(s1,s2,strlen(s2)))

/* to determine if a string is zero length */

#define ZEROLEN(str) (str[0]=='\0')

/* logical xor */

#define LXOR(x,y) (((x) && !(y)) || (!(x) && (y)))

/* square a value */

#define SQV(x) ((x)*(x))

/* ABS of a value */

#define MY_ABS(x) ((x) >= 0 ? (x) : -(x))

#endif	/* UTIL_UTIL_H */
